package blobovniczatree

import (
	"context"
	"encoding/hex"
	"path/filepath"
	"time"

	"git.frostfs.info/TrueCloudLab/frostfs-node/internal/logs"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/blobovnicza"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/blobstor/common"
	"git.frostfs.info/TrueCloudLab/frostfs-observability/tracing"
	"git.frostfs.info/TrueCloudLab/frostfs-sdk-go/client"
	"go.opentelemetry.io/otel/attribute"
	"go.opentelemetry.io/otel/trace"
	"go.uber.org/zap"
)

// Exists implements common.Storage.
func (b *Blobovniczas) Exists(ctx context.Context, prm common.ExistsPrm) (common.ExistsRes, error) {
	var (
		startedAt = time.Now()
		success   = false
		found     = false
	)
	defer func() {
		b.metrics.Exists(time.Since(startedAt), success, prm.StorageID != nil)
	}()

	ctx, span := tracing.StartSpanFromContext(ctx, "Blobovniczas.Exists",
		trace.WithAttributes(
			attribute.String("path", b.rootPath),
			attribute.String("address", prm.Address.EncodeToString()),
			attribute.String("storage_id", hex.EncodeToString(prm.StorageID)),
		))
	defer span.End()

	if prm.StorageID != nil {
		id := blobovnicza.NewIDFromBytes(prm.StorageID)
		blz, err := b.openBlobovnicza(id.String())
		if err != nil {
			return common.ExistsRes{}, err
		}

		exists, err := blz.Exists(ctx, prm.Address)
		return common.ExistsRes{Exists: exists}, err
	}

	activeCache := make(map[string]struct{})

	var gPrm blobovnicza.GetPrm
	gPrm.SetAddress(prm.Address)

	err := b.iterateSortedLeaves(ctx, &prm.Address, func(p string) (bool, error) {
		dirPath := filepath.Dir(p)

		_, ok := activeCache[dirPath]

		_, err := b.getObjectFromLevel(ctx, gPrm, p, !ok)
		if err != nil {
			if !client.IsErrObjectNotFound(err) {
				b.log.Debug(logs.BlobovniczatreeCouldNotGetObjectFromLevel,
					zap.String("level", p),
					zap.String("error", err.Error()))
			}
		}

		activeCache[dirPath] = struct{}{}
		found = err == nil
		return found, nil
	})

	success = err == nil
	return common.ExistsRes{Exists: found}, err
}
